//-----------------------------------------------------------------------------
//
// Interface code for Hypre solver.
//
// See https://github.com/LLNL/hypre/blob/master/src/examples/ex5.c
//
//-----------------------------------------------------------------------------

#include <stdio.h>
#include "mpi.h"

#include <math.h>
#include "_hypre_utilities.h"
#include "HYPRE_krylov.h"
#include "HYPRE.h"
#include "HYPRE_parcsr_ls.h"

//-----------------------------------------------------------------------------

// Assertion macro
//#define Assert(condition) {if ( !(condition) ) {                           \
//        fprintf(stderr, "Assertion error: \"%s\". At file %s, line %i.\n", \
//                #condition, __FILE__, __LINE__);                           \
//    exit(EXIT_FAILURE);                                                    \
//  }}
#define Assert(condition)

//-----------------------------------------------------------------------------

// Set relevant precisions.
// Always use double for linear solve, but data received from GTC
// may be single or double based on GTC build option.

typedef double Float_t;

#ifdef DOUBLE_PRECISION
typedef double GTC_float_t;
#else
typedef float GTC_float_t;
#endif

// Struct to hold GTC/Hypre solver state.

typedef struct {
  HYPRE_IJVector b;
  HYPRE_IJVector x;
  HYPRE_ParVector par_b;
  HYPRE_ParVector par_x;
  HYPRE_IJMatrix a;
  HYPRE_ParCSRMatrix parcsr_a;
  HYPRE_Solver precond;
  HYPRE_Solver solver;
  MPI_Comm comm;
  int comm_size;
  int comm_rank;
  Float_t* x_gathered;
  int* recvcounts;
  int* displs;
  Float_t* b_values;
  Float_t* x_values;
  int* row_indices;
  int n_global;
  int ilo;
  int ihi;
  int n_local;
} HypreGtcCommon;

// Global variable to hold solve state info,
// analogous to GTC Fortran common block.

HypreGtcCommon g_hypre_gtc_common2;

//-----------------------------------------------------------------------------
// Get base index in global coords for part of vector on this rank.

int hypre_vec_index(HypreGtcCommon* data, int proc) {
  return (data->n_global * (size_t)proc) / data->comm_size;
}

//-----------------------------------------------------------------------------
// Hypre basic initializations.

#ifdef __cplusplus
extern "C"
#endif
void hypre_gtc_initialize_(int* partd_comm_fortran) {

  // Select struct to use.

  HypreGtcCommon* const data = &g_hypre_gtc_common2;

  // Get the communicator to run a solve in.
  // Must convert from Fortran communicator to C communicator.

  MPI_Comm partd_comm = MPI_Comm_f2c(*partd_comm_fortran);

  MPI_Comm_dup(partd_comm, &data->comm);

  MPI_Comm_size(data->comm, &data->comm_size);
  MPI_Comm_rank(data->comm, &data->comm_rank);
}

//-----------------------------------------------------------------------------
// Terminations - undo effects of hypre_initialize_, hypre_create_.

#ifdef __cplusplus
extern "C"
#endif
void hypre_gtc_finalize_() {

  // Select struct to use.

  HypreGtcCommon* const data = &g_hypre_gtc_common2;

  free(data->x_gathered);
  free(data->recvcounts);
  free(data->displs);

  HYPRE_BoomerAMGDestroy(data->precond);
  HYPRE_ParCSRFlexGMRESDestroy(data->solver);

  HYPRE_IJVectorDestroy(data->b);
  HYPRE_IJVectorDestroy(data->x);
  free(data->row_indices);
  free(data->b_values);
  free(data->x_values);

  MPI_Comm_free(&data->comm);
}

//-----------------------------------------------------------------------------
// Set up vectors, matrix and solver.

#ifdef __cplusplus
extern "C"
#endif
void hypre_gtc_create_(int* n_global, int* lusers, int* luserp,
                       GTC_float_t* lusera) {

  // Select struct to use.

  HypreGtcCommon* const data = &g_hypre_gtc_common2;

  // The GTC partd communicator corresponds to the grid for a single
  // 2D slice of the torus.
  // Every MPI rank in this communicator contains a replicated copy
  // of the entire linear system for this communicator.
  // To parallelize this within the communicator, a block row
  // decomposition of the system is imposed, and each rank takes
  // its part of the linear system from the global system to do the solve.
  // At the end of the solve, each rank has part of the solution.
  // An MPI allgatherv is performed to give every rank the entire
  // solution vector, as GTC needs.

  //---
  // Set up vectors.
  //---

  // Get global and local vector sizes and local vector start/end coords.

  // NOTE: 32 bit integers being used here, may be a problem in the future.

  data->n_global = *n_global;

  data->ilo = hypre_vec_index(data, data->comm_rank);
  data->ihi = hypre_vec_index(data, data->comm_rank+1);

  data->n_local = data->ihi - data->ilo;

  // Allocate vectors: Hypre vectors and vector objects, vector data arrays.

  HYPRE_IJVectorCreate(data->comm, data->ilo, data->ihi-1, &data->b);
  HYPRE_IJVectorCreate(data->comm, data->ilo, data->ihi-1, &data->x);
  HYPRE_IJVectorSetObjectType(data->b, HYPRE_PARCSR);
  HYPRE_IJVectorSetObjectType(data->x, HYPRE_PARCSR);
  HYPRE_IJVectorInitialize(data->b);
  HYPRE_IJVectorInitialize(data->x);
  HYPRE_IJVectorAssemble(data->b);
  HYPRE_IJVectorAssemble(data->x);
  HYPRE_IJVectorGetObject(data->b, (void **)&(data->par_b));
  HYPRE_IJVectorGetObject(data->x, (void **)&(data->par_x));

  data->b_values = (Float_t*)malloc(data->n_local*sizeof(*(data->b_values)));
  data->x_values = (Float_t*)malloc(data->n_local*sizeof(*(data->x_values)));

  // Global element numbers of local vector arrays.

  data->row_indices = (int*)malloc(data->n_local*sizeof(*(data->row_indices)));
  for (int i=0; i<data->n_local; ++i) {
    data->row_indices[i] = data->ilo + i;
  }

  // Set up arrays needed for final vector gather.

  data->x_gathered = (Float_t*)malloc(data->n_global*
                                      sizeof(*(data->x_gathered)));
  data->recvcounts = (int*)malloc(data->comm_size*sizeof(*(data->recvcounts)));
  data->displs = (int*)malloc(data->comm_size*sizeof(*(data->displs)));

  for (int p=0; p<data->comm_size; ++p) {
    const int ilo = hypre_vec_index(data, p);
    const int ihi = hypre_vec_index(data, p+1);
    data->recvcounts[p] = ihi - ilo;
    data->displs[p] = ilo;
  }

  //---
  // Set up matrix.
  //---

  const int nnz_local = lusers[data->ihi] - lusers[data->ilo];

  int* row_lengths = (int*)malloc((data->n_local)*sizeof(*row_lengths));
  int* col_indices_global = (int*)malloc(nnz_local*
                                         sizeof(*col_indices_global));
  Float_t* values = (Float_t*)malloc(nnz_local*sizeof(*values));

  // Create required sparse matrix storage arrays.

  int i0 = 0;
  int j0 = 0;
  for (int i=data->ilo; i<data->ihi; ++i) {
    for (int j=lusers[i]; j<lusers[i+1]; ++j) {
      Assert(j0 < nnz_local);
      col_indices_global[j0] = luserp[j];
      values[j0] = (Float_t)(lusera[j]);
      Assert(col_indices_global[j0] >= 0);
      Assert(col_indices_global[j0] < data->n_global);
      ++j0;
    } //---j
    Assert(i0+1 < data->n_local+1);
    row_lengths[i0] = lusers[i+1] - lusers[i];
    Assert(row_lengths[i0] >= 0);
    ++i0;
  } //---i

  // Allocate and set matrix: Hypre matrix and matrix object.

  HYPRE_IJMatrixCreate(data->comm, data->ilo, data->ihi-1, data->ilo,
                       data->ihi-1, &data->a);
  HYPRE_IJMatrixSetObjectType(data->a, HYPRE_PARCSR);
  HYPRE_IJMatrixInitialize(data->a);
  HYPRE_IJMatrixSetValues(data->a, data->n_local, row_lengths,
                          data->row_indices, col_indices_global, values);
  HYPRE_IJMatrixAssemble(data->a);
  HYPRE_IJMatrixGetObject(data->a, (void**) &data->parcsr_a);

  free(row_lengths);
  free(col_indices_global);
  free(values);

  //---
  // Set up solver.
  //---

  HYPRE_ParCSRFlexGMRESCreate(data->comm, &data->solver);

  HYPRE_FlexGMRESSetKDim(data->solver, 50);
  HYPRE_FlexGMRESSetMaxIter(data->solver, 100);
  HYPRE_FlexGMRESSetTol(data->solver, 1e-6);
  //HYPRE_FlexGMRESSetPrintLevel(data->solver, 3);
  //HYPRE_FlexGMRESSetLogging(data->solver, 1);

  HYPRE_BoomerAMGCreate(&data->precond);

  HYPRE_BoomerAMGSetCoarsenType(data->precond, 6);
  //HYPRE_BoomerAMGSetOldDefault(data->precond);
  HYPRE_BoomerAMGSetRelaxType(data->precond, 6); /* Sym G.S./Jacobi hybrid */
  HYPRE_BoomerAMGSetNumSweeps(data->precond, 1);
  HYPRE_BoomerAMGSetTol(data->precond, 0.0);
  HYPRE_BoomerAMGSetMaxIter(data->precond, 1);
#ifdef _HYPRE_USES_GPU
  HYPRE_BoomerAMGSetRelaxType(data->precond, 18);
  HYPRE_BoomerAMGSetNumSweeps(data->precond , 2);
  HYPRE_BoomerAMGSetKeepTranspose(data->precond, 1);
  //HYPRE_BoomerAMGSetCoarsenType(data->precond, 8);
#endif

//int comm_rank_world;
//MPI_Comm_rank(MPI_COMM_WORLD, &comm_rank_world);
//if(comm_rank_world == 0)
//  HYPRE_BoomerAMGSetPrintLevel(data->precond, 3);

  HYPRE_FlexGMRESSetPrecond(data->solver,
                            (HYPRE_PtrToSolverFcn)HYPRE_BoomerAMGSolve,
                            (HYPRE_PtrToSolverFcn)HYPRE_BoomerAMGSetup,
                            data->precond);

  HYPRE_ParCSRFlexGMRESSetup(data->solver, data->parcsr_a, data->par_b,
                             data->par_x);
}

//-----------------------------------------------------------------------------
// Perform a single linear system solve.

#ifdef __cplusplus
extern "C"
#endif
void hypre_gtc_solver_(GTC_float_t* luserx, GTC_float_t* luserb) {

  // Select struct to use.

  HypreGtcCommon* const data = &g_hypre_gtc_common2;

  // Extract portion of global rhs, initial guess to use for solve on this rank.

  GTC_float_t* b_values = luserb + data->ilo;
  GTC_float_t* x_values = luserx + data->ilo;

  for (int i=0; i<data->n_local; ++i) {
    data->b_values[i] = (Float_t)(b_values[i]);
    data->x_values[i] = (Float_t)(x_values[i]);
  }

  // Set HYPRE vectors.

  HYPRE_IJVectorInitialize(data->b);
  HYPRE_IJVectorInitialize(data->x);
  HYPRE_IJVectorSetValues(data->b, data->n_local, data->row_indices,
                          data->b_values);
  HYPRE_IJVectorSetValues(data->x, data->n_local, data->row_indices,
                          data->x_values);
  HYPRE_IJVectorAssemble(data->b);
  HYPRE_IJVectorAssemble(data->x);

  // Perform solve.

  HYPRE_ParCSRFlexGMRESSolve(data->solver, data->parcsr_a, data->par_b,
                             data->par_x);

  // Get solution vector.

  HYPRE_IJVectorGetValues(data->x, data->n_local, data->row_indices,
                          data->x_values);

  // Distribute each rank's portion of solution to all ranks.

  MPI_Allgatherv(data->x_values, data->n_local,
                 sizeof(Float_t) == 8 ? MPI_DOUBLE : MPI_FLOAT,
                 data->x_gathered, data->recvcounts, data->displs,
                 sizeof(Float_t) == 8 ? MPI_DOUBLE : MPI_FLOAT, data->comm);

  for (int i=0; i<data->n_global; ++i) {
    luserx[i] = (GTC_float_t)(data->x_gathered[i]);
  }
}

//-----------------------------------------------------------------------------



